# v0.03 n3617400
# uniden homepatrol-1 firmware decrypter

import os
import sys

def scn_parse_string(s):
	address_len = (2,2,3,4,2,4,3,2)[int(s[1])]
	bs = list(bytes.fromhex(s[2:].rstrip()))
	if sum(bs[:-1]) & 0xff != bs[-1] ^ 0xff:
		raise Exception('crc', s)
	return (int(s[4:4+address_len*2],16), bs[1+address_len:-1])

def scn_parse_file(file_name):
	header = ()
	data = []
	for line in open(file_name,'r'):
		r = int(line[1])
		if r == 0:
			header = scn_parse_string(line)
		elif r == 1 or r == 2 or r == 3:
			d = scn_parse_string(line)
			if not len(data):
				data.append(d)
			else:
				last_index = len(data)-1
				addr = data[last_index][0] + len(data[last_index][1])
				if d[0] != addr:
					data.append(d)
				else:
					data[last_index][1].extend(d[1])
		elif r == 8:
			pass
		else:
			raise Exception('unknown scn type', r)
	return header, data

def decrypt(data):
	dkeys = {0:0xab, 1:0xbe, 2:0xcd, 3:0xca, 4:0xcd, 5:0xa6, 6:0xe0, 519:0xe0}
	for block in data:
		for i in range(0, len(block[1])):
			block[1][i] = (block[1][i] + dkeys[(block[0]+i) >> 20]) & 0xff

def main():
	print('uft by n3617400')
	argc = sys.argv[1:]
	if not argc:
		print('usage: uft.py firmware_file.scn')
		return
	dir = os.path.splitext(argc[0])[0]
	if os.path.exists(dir):
		raise Exception('wrong: '+dir)
		return
	os.mkdir(dir)

	header, data = scn_parse_file(argc[0])
	decrypt(data)

	for block in data:
		name = os.path.join(dir,'{:08x}_{:08x}.bin'.format(block[0],len(block[1])))
		with open(name,'wb') as file:
			file.write(bytearray(block[1]))
	print ('see dir: '+dir)

if __name__ == "__main__":
	main()


